--  file  : font.e - font library
--  author: jiri babor
--  email : jbabor@paradise.net.nz

--  date  : 98-12-04
--  ver   : 4.40
--      expanded set_parameter() routine: now it allows change of
--      entire sequence of parameters, not just individual parameters
--      otherwise cosmetic changes only


--  new features:
--      faster, very *fast* with plain foreground and colored background
--      foreground texturing option
--      8x16 ROM char set as default
--      font attributes: bold, shadow, italic & underline
--      vertical text
--      global routines only: *no* global variables any more
--      each font now has its own private copy of parameters
--      'on-the-fly' change of foreground color and/or attributes
--      multi-line support (imbedded new line chars)
--      text justification: left, right and centre
--      images() function: fast generation of rectangular text images
--      extra line spacing font parameter
--      reset_font() procedure
--      new (changed) defaults
--      prompt() function (formerly fgets) with additional width parameter,
--          scroll, xor cursor and much better edit facilities

--  font file format -----------------------------------------------------------
--      byte 1: ascii code of the first char
--      byte 2: ascii code of the last char
--      byte 3: width in bytes
--      byte 4: height in pixels
--      byte 5: version                           unused
--      byte 6: baseline height
--      next block of bytes: actual char widths
--      next block of bytes: simple char bitmaps, left-to-right, top-to-bottom

--  naming conventions ---------------------------------------------------------
--      All global constants are in upper case, everything else is in
--      lower only! No global variables are *directly* accessible. Names
--      of global functions are usually nouns, with the sole exception
--      of getx & gety pair. Names of all global procedures are verbs,
--      or combinations starting with the word 'set'.
--include Rncimage.e
include graphics.e          -- color constants etc
include machine.e           -- dos_interrupt()
include wildcard.e          -- lower()
tick_rate(100)
constant
    DEFAULT_PARAMS=         -- assigned to each loaded font
       {1,  -- 1: boldface offset
	1,  -- 2: shadow offset: right
	1,  -- 3: shadow offset: down
	0,  -- 4: extra kerning space width
	1,  -- 5: underline thickness
	0}  -- 6: extra line spacing

global constant
    CLEAR       =-1,    -- transparent background color
    ROM         =1,     -- default font

    -- attributes are 'additive', e.g.: BOLD+UNDERLINE=1+4=5
    NONE        =0,     -- plain
    BOLD        =1,
    ITALIC      =2,
    SHADOW      =4,
    UNDERLINE   =8,

    -- (multi-)line justification
    LEFT        =0,
    CENTRE      =1,
    RIGHT       =2,

    -- line orientation
    HORIZONTAL  =0,
    VERTICAL    =1

object
    fc              -- text foreground color or 2-d pattern (tile)

sequence
    bheights,       -- baseline heights of all loaded fonts
    font,           -- current font
    fonts,          -- all loaded fonts
    heights,        -- heights of all loaded fonts
    ilo,            -- italic line offset
    names,          -- font filenames
    pa,             -- current font parameters
    params,         -- sequence of all font parameters
    wi,             -- char widths, current font (in pixels)
    widths          -- char widths of all loaded fonts (in pixels)

integer
    at,             -- current combined attributes
    bc,             -- background color
    bh,             -- baseline height of current font
    bo,             -- boldface offset
    f,              -- current font
    h,              -- nominal height of current font
    italic,         -- flag
    ju,             -- text justification (see global constants above)
    orient,         -- line orientation, horizontal or vertical
    sc,             -- shadow color
    x,              -- text pointer: left edge char space
    y               -- text pointer: top  edge char space

function rtrim(sequence s)
    -- trim (discard) trailing zeros of sequence s
    integer i
    i=length(s)
    while i do
	if s[i] then exit end if
	i=i-1
    end while
    return s[1..i]
end function

procedure xor_cursor(integer x, integer y, integer h)
    -- xor cursor: vertical line
    for i=0 to h-1 do
	pixel(xor_bits(get_pixel({x,y+i}),BRIGHT_WHITE),{x,y+i})
    end for
end procedure

procedure error(sequence s)
    object junk
    junk=graphics_mode(-1)
    puts(1, s & "\n")
    abort(1)
end procedure

function loaded(sequence name)
    -- returns allocated handle if font already loaded, else zero
    integer i,f
    f=0   i=0
    while i<length(names) do
	i=i+1
	if compare(name,names[i])=0 then
	    f=i
	    exit
	end if
    end while
    return f
end function

global procedure align(integer previous_font)
    -- align current font with previous_font
    if previous_font>0 and previous_font<=length(bheights) then
	if orient then      -- VERTICAL
	    x=x+bheights[previous_font]-bh
	else                -- HORIZONTAL
	    y=y+bheights[previous_font]-bh
	end if
    end if
end procedure

global function attributes()
    -- return current text attributes
    return at
end function

global procedure set_attributes(integer a)
    -- set text attributes
    at=a
    bo=pa[1]*and_bits(at,BOLD)      -- new bold offset
    italic=and_bits(at,ITALIC)>0    -- new italic flag
end procedure

global function height()
    -- return nominal height of the current font in pixels
    return h
end function

global function width(integer c)
    -- return width of char c using current font, attributes and
    -- kerning, in pixels
    return wi[c]+bo+pa[4]
end function

global procedure reset_font()
    -- reset entire environment of current font to defaults
    x=0                 -- left edge of char space
    y=0                 -- top edge of char space
    at=0                -- font attributes, none
    bc=CLEAR            -- transparent background color
    fc=BRIGHT_WHITE     -- foreground text color
    sc=GRAY             -- shadow color
    ju=LEFT             -- default justification
    pa=DEFAULT_PARAMS   -- font parameters
    bo=0                -- bold offset, in pixels
    italic=0            -- flag only
    orient=HORIZONTAL   -- line orientation
end procedure

global procedure select_font(integer fh)
    -- select font with handle fh, make it current
    if fh>0 and fh<=length(fonts) and f!=fh then
	params[f]=pa                -- save current parameters
	f=fh
	font=fonts[f]
	wi=widths[f]
	h=heights[f]
	bh=bheights[f]
	pa=params[f]                -- retrieve saved parameters
	bo=pa[1]*and_bits(at,BOLD)  -- new bold offset
	ilo=repeat(0,h)
	for i=1 to h do ilo[i]=floor((h-i)*.25) end for
    end if
end procedure

global function load_font(sequence file_name)
    -- load font file into memory and make it available for selection
    object junk
    integer n,fn,h,byte_width,fc,first_char,last_char,len,bh
    sequence row_bits,chars,cs,cbs,w

    file_name=lower(file_name)
    n=loaded(file_name)
    if n then return n  -- if font already loaded, do not do it again
    else
	fn=open(file_name, "rb")
	if fn=-1 then
	    error("Font load error: " & file_name & " not found !")
	end if
	names=append(names,file_name)
	first_char=getc(fn)
	last_char=getc(fn)
	byte_width=getc(fn)
	h=getc(fn)                  -- nominal height
	junk=getc(fn)               -- version number, unused
	bh=getc(fn)                 -- baseline height
	if bh=0 then bh=h end if    -- emergency solution for pre 3.xx fonts

	-- get widths
	w=repeat(0,255)
	if first_char=0 then        -- skip zero char
	    fc=1
	    junk=getc(fn)           -- dump
	else fc=first_char
	end if
	for i=fc to last_char do
	    w[i]=getc(fn)
	end for

	-- get char images
	chars=repeat({},255)
	len=h*byte_width
	if first_char=0 then        -- again skip zero char
	    fc=1
	    for i=1 to len do       -- dump
		junk=getc(fn)
	    end for
	end if
	for i=fc to last_char do
	    cbs=repeat(0,len)       -- char byte sequence
	    cs={}                   -- char sequence of row bit sequences
	    for j=1 to len do
		cbs[j]=getc(fn)
	    end for
	    cbs=rtrim(cbs)          -- remove trailing (bottom) bytes
	    row_bits={}
	    for j=1 to length(cbs) do
		row_bits=row_bits & (and_bits(cbs[j],{128,64,32,16,8,4,2,1}) and 1)
		if remainder(j,byte_width)=0 then
		    cs=append(cs,rtrim(row_bits))
		    row_bits={}
		end if
	    end for
	    if length(row_bits) then
		cs=append(cs,rtrim(row_bits))
	    end if
	    chars[i]=cs
	end for
	close(fn)
	fonts=append(fonts,chars)
		widths=append(widths, w)
	heights=heights & h
	bheights=bheights & bh
	params=append(params,DEFAULT_PARAMS)
	return length(fonts)
    end if
end function  -- load_font

procedure load_rom_font()
    sequence
	regs,   -- list of register values
	chars,  -- bitmaps od all chars
	cbm,    -- char bitmap
	b,cb    -- char bytes
    integer
	a,      -- font pointer (starting address)
	len
    regs = repeat(0,10)
    regs[REG_AX]=#1130
    regs[REG_BX]=#600
    regs=dos_interrupt(#10,regs)
    a=#10*regs[REG_ES]+regs[REG_BP]+16  -- bypass zero char
    chars={}
    for i=1 to 255 do                   -- bypass zero char
	cbm={}
	cb=rtrim(peek({a,16}))          -- char bytes, right trimmed
	len=length(cb)
	for j=1 to length(cb) do
	    b=and_bits(cb[j],{128,64,32,16,8,4,2,1}) and 1
	    cbm=append(cbm,rtrim(b))
	end for
	a=a+16
	chars=append(chars,cbm)
    end for

    -- initialize fonts
    f=ROM
    names={"rom_font"}
    fonts={chars}
    widths={repeat(8,255)}
    heights={16}
    bheights={12}
    params={DEFAULT_PARAMS}

    -- select rom font as default
    font=fonts[f]
    wi=widths[f]
    h=16
    pa=DEFAULT_PARAMS
    ilo=repeat(0,h)
    for i=1 to h do ilo[i]=floor((h-i)*.25) end for
end procedure -- load_rom_font

global function lengths(sequence text)
    -- return sequence of lengths (in pixels) of string text
    -- using current font, attributes and parameters
    object len
    integer a,b,c,l,w,it
    len={}   l=0
    b=bo                                -- current bold offset
    a=at                                -- current attributes
    it=italic                           -- current italic flag
    for i=1 to length(text) do
	c=text[i]
	if c<=0 then
	    a=-c
	    b=pa[1]*and_bits(a,BOLD)    -- new bold offset
	    it=and_bits(a,ITALIC)>0     -- new italic flag
	elsif c=10 then                 -- new line
	    if l and it then            -- extra italic length of last char
		l=l+ilo[1]
	    end if
	    len=len & l                 -- append length of previous line
	    l=0                         -- starting length of new line
	elsif c<256 then
	    w=wi[c]
	    if w then l=l+w+b+pa[4] end if
	end if
    end for
    if l and it then                    -- extra italic length of last char
	l=l+ilo[1]
    end if
    return len & l                      -- append also length of last line
end function   -- lengths

procedure char(integer c)
    sequence cs,row,pix
    integer n,u,v,us,nr,nc,w
    cs=font[c]
    w=wi[c]
    if w then
	w=w+bo+pa[4]            -- current width of char
	if and_bits(at,SHADOW) then
	    v=0                 -- local vertical coordinate
	    for i=1 to length(cs) do
		row=cs[i]
		u=ilo[i]*italic -- local horizontal coordinate
		n=0             -- counter: adjacent bits
		for j=1 to length(row) do
		    if row[j] then
			if not n then us=u end if
			n=n+1
		    elsif n then
			pixel(repeat(sc,n+bo),{x+us+pa[2],y+v+pa[3]})
			n=0
		    end if
		    u=u+1
		end for
		if n then
		    pixel(repeat(sc,n+bo),{x+us+pa[2],y+v+pa[3]})
		end if
		v=v+1
	    end for
	    if and_bits(at,UNDERLINE) then
		for i=1 to pa[5] do
		    pixel(repeat(sc,w),{x+pa[2],y+h-i+pa[3]})
		end for
	    end if
	end if
	v=0
	if integer(fc) then                 -- single color foreground
	    for i=1 to length(cs) do
		row=cs[i]
		u=ilo[i]*italic             -- local horizontal coordinate
		n=0
		for j=1 to length(row) do
		    if row[j] then
			if not n then us=u end if
			n=n+1
		    elsif n then
			pixel(repeat(fc,n+bo),{x+us,y+v})
			n=0
		    end if
		    u=u+1
		end for
		if n then pixel(repeat(fc,n+bo),{x+us,y+v}) end if
		v=v+1
	    end for
	    if and_bits(at,UNDERLINE) then
		for i=1 to pa[5] do
		    pixel(repeat(fc,w),{x,y+h-i})
		end for
	    end if
	else                                -- tiled foreground
	    nr=length(fc)                   -- nuber of tile rows (mod y)
	    nc=length(fc[1])                -- nuber of tile columns (mod x)
	    for r=1 to length(cs) do
		u=ilo[r]*italic             -- local horizontal coordinate
		row=cs[r]
		n=0                         -- not a count, just a flag
		pix={}
		for col=1 to length(row) do
		    if row[col] then
			if not n then us=u end if
			n=1
			pix=pix & fc[remainder(v,nr)+1][remainder(u,nc)+1]
		    elsif n then
			for i=1 to bo do
			    pix=pix & fc[remainder(v,nr)+1][remainder(u+i,nc)+1]
			end for
			pixel(pix,{x+us,y+v})
			n=0
			pix={}
		    end if
		    u=u+1
		end for
		if n then
		    for i=1 to bo do
			pix=pix & fc[remainder(v,nr)+1][remainder(u+i,nc)+1]
		    end for
		    pixel(pix,{x+us,y+v})
		end if
		v=v+1
	    end for
	    if and_bits(at,UNDERLINE) then
		for j=h-pa[5] to h-1 do
		    pix=repeat(0,w)
		    for i=0 to w-1 do
			pix[i+1]=fc[remainder(j,nr)+1][remainder(i,nc)+1]
		    end for
		    pixel(pix,{x,y+j})
		end for
	    end if
	end if
	x=x+w
    end if
end procedure -- char

global function images(sequence s)
    -- return sequence of rectangular text images corresponding to (multi-line)
    -- string s, and using current font, attributes, colors and parameters
    sequence cs,ims,len,line,row
    integer c,lx,m,n,u,w,nc,nr
    if bc>=0 then                       -- opaque background
	ims={}
	len=lengths(s)
	n=1                             -- first line (image)
	line=repeat(repeat(bc,len[n]),h)
	lx=1                            -- char horizontal line position
	for i=1 to length(s) do
	    c=s[i]
	    if c<=0 then set_attributes(-c)
	    elsif c>255 then fc=c-256   -- new text color
	    elsif c=10 then             -- append current and start new line
		ims=append(ims,line)
		n=n+1
		line=repeat(repeat(bc,len[n]),h)
		lx=1                    -- char horizontal line position
	    else                        -- printable character
		cs=font[c]
		w=wi[c]
		if w then
		    w=w+bo+pa[4]
		    if integer(fc) then -- single color foreground
			for r=1 to length(cs) do
			    row=cs[r]
			    u=lx+ilo[r]*italic
			    m=0     -- flag: set when previous pixel on
			    for j=1 to length(row) do
				if row[j] then
				    line[r][u]=fc
				    m=1
				elsif m then
				    if bo then
					for k=1 to bo do
					    line[r][u+k-1]=fc
					end for
				    end if
				    m=0
				end if
				u=u+1
			    end for
			    if m and bo then
				for k=1 to bo do
				    line[r][u+k-1]=fc
				end for
			    end if
			end for
			if and_bits(at,UNDERLINE) then
			    for k=1 to pa[5] do
				line[h-k+1][lx..lx+w-1]=fc
			    end for
			end if
		    else                -- tiled foreground
			nr=length(fc)       -- nuber of tile rows (mod y)
			nc=length(fc[1])    -- nuber of tile columns (mod x)
			for r=1 to length(cs) do
			    row=cs[r]
			    u=lx+ilo[r]*italic
			    m=0     -- flag: set when previous pixel on
			    for j=1 to length(row) do
				if row[j] then
				    line[r][u]=fc[remainder(r-1,nr)+1][remainder(u-1,nc)+1]
				    m=1
				elsif m then
				    if bo then
					for k=1 to bo do
					    line[r][u+k-1]=fc[remainder(r-1,nr)+1][remainder(u-2+k,nc)+1]
					end for
				    end if
				    m=0
				end if
				u=u+1
			    end for
			    if m and bo then
				for k=1 to bo do
				    line[r][u+k-1]=fc[remainder(r-1,nr)+1][remainder(u-2+k,nc)+1]
				end for
			    end if
			end for
			if and_bits(at,UNDERLINE) then
			    for k=0 to pa[5]-1 do
				for j=0 to w-1 do
				    line[h-k][lx+j]=fc[remainder(h-k-1,nr)+1][remainder(lx+j-1,nc)+1]
				end for
			    end for
			end if
		    end if
		    lx=lx+w
		end if
	    end if
	end for
	return  append(ims,line)
    else error("Images error: opaque background not specified.")
    end if
end function -- images

global procedure write_horizontal(sequence s)
    -- display text s
    sequence cs,len,line,row
    integer c,m,n,ox,u,lx,w,nc,nr
    ox=x                                -- original x coordinate
    len=lengths(s)
    n=1                                 -- first line
    if ju=CENTRE then
	x=floor(ox-(len[n]+1)/2)
    elsif ju=RIGHT then
	x=ox-len[n]
    end if
    if bc=CLEAR then                    -- transparent background - slower mode
	for i=1 to length(s) do
	    c=s[i]
	    if c<=0 then set_attributes(-c)
	    elsif c>255 then fc=c-256   -- new text color
	    elsif c=10 then             -- new line
		if ju=CENTRE then
		    n=n+1
		    x=floor(ox-(len[n]+1)/2)
		elsif ju=RIGHT then
		    n=n+1
		    x=ox-len[n]
		else
		    x=ox
		end if
		y=y+h+pa[6]
	    else char(c)                -- printable character
	    end if
	end for
    else            -- fast mode, opaque background: print whole lines
	line=repeat(repeat(bc,len[1]),h)
	lx=1                            -- char horizontal line position
	for i=1 to length(s) do
	    c=s[i]
	    if c<=0 then set_attributes(-c)
	    elsif c>255 then fc=c-256   -- new text color
	    elsif c=10 then             -- new line
		-- display line
		for j=1 to length(line) do
		    pixel(line[j],{x,y+j-1})
		end for
		n=n+1
		if ju=CENTRE then
		    x=floor(ox-(len[n]+1)/2)
		elsif ju=RIGHT then
		    x=ox-len[n]
		end if
		y=y+h+pa[6]
		line=repeat(repeat(bc,len[n]),h)
		lx=1                    -- char horizontal line position
	    else                        -- printable character
		cs=font[c]
		w=wi[c]
		if w then
		    w=w+bo+pa[4]
		    if integer(fc) then -- single color foreground
			for r=1 to length(cs) do
			    row=cs[r]
			    u=lx+ilo[r]*italic
			    m=0     -- flag: set when previous pixel on
			    for j=1 to length(row) do
				if row[j] then
				    line[r][u]=fc
				    m=1
				elsif m then
				    if bo then
					for k=1 to bo do
					    line[r][u+k-1]=fc
					end for
				    end if
				    m=0
				end if
				u=u+1
			    end for
			    if m and bo then
				for k=1 to bo do
				    line[r][u+k-1]=fc
				end for
			    end if
			end for
			if and_bits(at,UNDERLINE) then
			    for k=1 to pa[5] do
				line[h-k+1][lx..lx+w-1]=fc
			    end for
			end if
		    else                    -- tiled foreground
			nr=length(fc)       -- nuber of tile rows (mod y)
			nc=length(fc[1])    -- nuber of tile columns (mod x)
			for r=1 to length(cs) do
			    row=cs[r]
			    u=lx+ilo[r]*italic
			    m=0             -- flag: set when previous pixel on
			    for j=1 to length(row) do
				if row[j] then
				    line[r][u]=fc[remainder(r-1,nr)+1][remainder(u-1,nc)+1]
				    m=1
				elsif m then
				    if bo then
					for k=1 to bo do
					    line[r][u+k-1]=fc[remainder(r-1,nr)+1][remainder(u-2+k,nc)+1]
					end for
				    end if
				    m=0
				end if
				u=u+1
			    end for
			    if m and bo then
				for k=1 to bo do
				    line[r][u+k-1]=fc[remainder(r-1,nr)+1][remainder(u-2+k,nc)+1]
				end for
			    end if
			end for
			if and_bits(at,UNDERLINE) then
			    for k=0 to pa[5]-1 do
				for j=0 to w-1 do
				    line[h-k][lx+j]=fc[remainder(h-k-1,nr)+1][remainder(lx+j-1,nc)+1]
				end for
			    end for
			end if
		    end if
		    lx=lx+w
		end if
	    end if
	end for
	-- display last line
	for j=1 to length(line) do
	    pixel(line[j],{x,y+j-1})
	end for
	x=x+len[n]
    end if
end procedure

procedure vchar(integer c)
    -- char display in vertical mode
    sequence cs,row
    integer n,u,v,nr,nc,w
    cs=font[c]
    w=wi[c]
    if w then
	w=w+bo+pa[4]                -- current width of char
	if and_bits(at,SHADOW) then
	    u=0                     -- local horizontal coordinate
	    for i=1 to length(cs) do
		row=cs[i]
		v=ilo[i]*italic     -- local vertical coordinate
		n=0                 -- flag: previous bit was set
		for j=1 to length(row) do
		    if row[j] then
			pixel(sc,{x+u+pa[2],y-v-pa[3]})
			n=1
		    elsif n and bo then
			n=0
			for k=1 to bo do
			    pixel(sc,{x+u+pa[2],y-v-k-pa[3]})
			end for
		    end if
		    v=v+1
		end for
		if n and bo then
		    v=v-1
		    for k=1 to bo do
			pixel(sc,{x+u+pa[2],y-v-k-pa[3]})
		    end for
		end if
		u=u+1
	    end for
	    if and_bits(at,UNDERLINE) then
		for i=1 to w do
		    for j=1 to pa[5] do
			pixel(sc,{x+h-j+pa[2],y-i+1-pa[3]})
		    end for
		end for
	    end if
	end if
	u=0
	if sequence(fc) then                -- tiled foreground
	    nr=length(fc)                   -- nuber of tile rows (mod y)
	    nc=length(fc[1])                -- nuber of tile columns (mod x)
	end if
	for i=1 to length(cs) do
	    row=cs[i]
	    v=ilo[i]*italic     -- local vertical coordinate
	    n=0                 -- flag: previous bit was set
	    for j=1 to length(row) do
		if row[j] then
		    if integer(fc) then pixel(fc,{x+u,y-v})
		    else pixel(fc[remainder(v,nr)+1][remainder(u,nc)+1],{x+u,y-v})
		    end if
		    n=1
		elsif n and bo then
		    n=0
		    for k=1 to bo do
			if integer(fc) then pixel(fc,{x+u,y-v-k+1})
			else pixel(fc[remainder(v+k-1,nr)+1][remainder(u,nc)+1],{x+u,y-v-k+1})
			end if
		    end for
		end if
		v=v+1
	    end for
	    if n and bo then
		for k=1 to bo do
		    if integer(fc) then pixel(fc,{x+u,y-v-k+1})
		    else pixel(fc[remainder(v+k-1,nr)+1][remainder(u,nc)+1],{x+u,y-v-k+1})
		    end if
		end for
	    end if
	    u=u+1
	end for
	if and_bits(at,UNDERLINE) then
	    for i=1 to w do
		for j=1 to pa[5] do
		    if integer(fc) then pixel(fc,{x+h-j,y-i+1})
		    else pixel(fc[remainder(i-1,nr)+1][remainder(h-j,nc)+1],{x+h-j,y-i+1})
		    end if
		end for
	    end for
	end if
	y=y-w
    end if
end procedure -- vchar

procedure write_vertical(sequence s)
    -- display text, vertically
    sequence len
    integer c,n,oy
    oy=y                            -- original x coordinate
    len=lengths(s)
    n=1                             -- first line
    if ju=CENTRE then
	y=floor(oy+(len[n]+1)/2)
    elsif ju=RIGHT then
	y=oy+len[n]
    end if
    for i=1 to length(s) do
	c=s[i]
	if c<=0 then set_attributes(-c)
	elsif c>255 then fc=c-256   -- new text color
	elsif c=10 then             -- new line
	    n=n+1
	    if ju=CENTRE then
		y=floor(oy+(len[n]+1)/2)
	    elsif ju=RIGHT then
		y=oy+len[n]
	    else
		y=oy
	    end if
	    x=x+h+pa[6]
	else vchar(c)                -- printable character
	end if
    end for
end procedure -- write_vertical

global procedure write(sequence s)
    if orient then write_vertical(s)
    else write_horizontal(s)
    end if
end procedure

function strip(sequence s)
    -- return string s stripped of all formatting characters & \n,\r, \t
    sequence ss
    integer c
    ss={}
    for i=1 to length(s) do
	c=s[i]
	if c>0 and c<256 and not find(c,"\n\r\t") then ss=ss&c end if
    end for
    return ss
end function

function slen(sequence text)
    -- simplified version of lengths()
    -- return length of already stripped text, in pixels
    integer len,w
    len=0
    for i=1 to length(text) do
	w=wi[text[i]]
	if w then
	    len=len+w+bo+pa[4]
	end if
    end for
    return len
end function -- slen

global function prompt(sequence s,integer w)
    -- prompted screen input of a string with editing
    -- s is the initial prompt string, w is the max field width, in pixels
    -- any formatting chars in the prompt string s are ignored (stripped)
    -- current attributes and line justification are also ignored
    -- return two element sequence {key, str} when one of the following
    -- keys is pressed: tab, shift-tab, enter, esc, up or down arrow
    -- str is the current state of the prompt string, its length is *not*
    -- restricted by the field width, the display is fully scrollable
    atom t
    integer
	att,    -- saved attributes
	bco,    -- saved background color
	jus,    -- saved justification
	ori,    -- saved line orientation
	temp,
	p,      -- position, general
	cc,     -- cursor color
	key,    -- pressed key
	dx,dy,
	pc,     -- paper color
	flag,   -- cursor flag, drawn/erased
	len,    -- length, in pixels
	l,      -- current string length, number of chars
	ci,     -- cursor index, index of char under it
	fvc,    -- index of first visible char
	sx      -- starting x, left edge of box

    att=at                  -- save current attributes
    at=NONE                 -- current attributes are ignored
    bco=bc                  -- save current background
    bo=0                    -- bold offset
    italic=0                -- italic flag
    jus=ju                  -- save current justification
    ju=LEFT
    ori=orient
    orient=HORIZONTAL
    s=strip(s)              --  strip all formatting characters & \n,\r, \t
    dy=h
    if bc=CLEAR then
	pc=get_pixel({x,y})
    else
	pc=bc
	bc=CLEAR
    end if
    sx=x
    l=length(s)
    fvc=1
    ci=l+1
    while slen(s[fvc..l])>w do fvc=fvc+1 end while
    write(s[fvc..l])

    -- blinking xor cursor
    cc=BRIGHT_WHITE
    while 1 do
	t=time()
	xor_cursor(x,y,h-3)
	flag=1                      -- cursor drawn
	while 1 do
	    if time()-t > 0.2 then
		t=time()
		xor_cursor(x,y,h-3)
		flag=not flag  end if
	    key=get_key()
	    if key!=-1 then
		-- erase current cursor if necessary
		if flag then
		    xor_cursor(x,y,h-3)
		end if
		exit end if end while
	if key=331 then      -- left arrow
	    if ci>1 then
		ci=ci-1
		if ci<fvc then  -- at left edge of box - scroll
		    fvc=fvc-1
		    len=slen(s[ci..l])
		    if len>w then
			p=l
			while len>w do
			    p=p-1
			    len=slen(s[ci..p]) end while
			polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+dy},{sx,y+dy}})
			x=sx
			write(s[ci..p])
			x=sx
		    else
			polygon(pc,1,{{sx,y},{sx+len-1,y},{sx+len-1,y+dy},{sx,y+dy}})
			x=sx
			write(s[ci..l])
			x=sx end if
		else x=x-slen({s[ci]}) end if end if
	elsif key=333 then      -- right arrow
	    if ci<=l then
		len=slen(s[fvc..ci])
		if len>w then   -- right box edge
		    fvc=fvc+1
		    len=slen(s[fvc..ci])
		    while len>w do
			fvc=fvc+1
			len=slen(s[fvc..ci]) end while
		    x=sx
		    polygon(pc,1,{{x,y},{x+w-1,y},{x+w-1,y+dy},{x,y+dy}})
		    write(s[fvc..ci])
		    if ci<l then
			if len+slen({s[ci+1]})<=w then
			    write({s[ci+1]}) end if end if
		    x=sx+len
		else x=x+slen({s[ci]}) end if
		ci=ci+1 end if
	elsif key=327 then  -- home key
	    if fvc>1 then
		p=0
		while p<=l do
		    len=slen(s[1..p])
		    if slen(s[1..p])>w then
			p=p-1
			exit
		    else p=p+1 end if end while
		polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+dy},{sx,y+dy}})
		x=sx
		write(s[1..p]) end if
	    ci=1
	    x=sx
	    fvc=1
	elsif key=335 then  -- end key
	    if ci!=l+1 then
		fvc=1
		ci=l+1
		while slen(s[fvc..l])>w do fvc=fvc+1 end while
		polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+dy},{sx,y+dy}})
		x=sx
		write(s[fvc..l]) end if
	elsif key=339 then  -- delete
	    if ci and ci<=l then
		temp=x
		len=slen(s[ci+1..l])
		if len>w-x+sx then
		    p=l
		    while len>w-x+sx do
			p=p-1
			len=slen(s[ci+1..p]) end while
		    polygon(pc,1,{{x,y},{sx+w-1,y},{sx+w-1,y+dy},{x,y+dy}})
		    write(s[ci+1..p])
		else
		    len=len+slen({s[ci]})
		    if len>w-x+sx then len=w-x+sx end if
		    polygon(pc,1,{{x,y},{x+len-1,y},{x+len-1,y+dy},{x,y+dy}})
		    write(s[ci+1..l]) end if
		x=temp
		s=s[1..ci-1]&s[ci+1..l]
		l=l-1 end if
	elsif key=8 then    -- backspace
	    if ci>1 then
		ci=ci-1
		l=l-1
		if ci>l then    -- at end of string
		    if ci<fvc then  -- at left edge of box
			fvc=fvc-1
		    else
			dx=slen({s[ci]})
			x=x-dx
			-- and erase the last char
			polygon(pc,1,{{x,y},{x+dx-1,y},{x+dx-1,y+dy},
			{x,y+dy}}) end if
		    s=s[1..l]
		else    -- inside string
		    if ci<fvc then fvc=fvc-1    -- at left edge of box
		    else
			dx=slen({s[ci]})
			x=x-dx
			temp=x
			len=slen(s[ci+1..l+1])
			if len>w-x+sx then
			    p=l
			    len=slen(s[ci+1..p])
			    while len>w-x+sx do
				p=p-1
				len=slen(s[ci+1..p]) end while
			    polygon(pc,1,{{x,y},{sx+w-1,y},{sx+w-1,y+dy},{x,y+dy}})
			    write(s[ci+1..p])
			    x=temp
			else
			    polygon(pc,1,{{x,y},{x+len+dx-1,y},{x+len+dx-1,y+dy},{x,y+dy}})
			    write(s[ci+1..l+1])
			    x=temp end if end if
		    s=s[1..ci-1]&s[ci+1..l+1] end if end if
	elsif find(key,{9,13,27,271,328,336}) then exit
	elsif key>0 and key<256 then
	    if ci>l then        -- at end of string
		s=s&key
		l=l+1
		ci=ci+1
		len=slen(s[fvc..l])
		if len<=w then char(key)
		else            -- past right edge of box - scroll left
		    while len>w do
			fvc=fvc+1
			len=slen(s[fvc..l]) end while
		    polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+dy},{sx,y+dy}})
		    x=sx
		    write(s[fvc..l]) end if
	    else
		s=s[1..ci-1]&key&s[ci..l]
		l=l+1
		temp=x
		len=slen(s[fvc..l])
		if len<=w then
		    polygon(pc,1,{{x,y},{sx+w-1,y},{sx+w-1,y+dy},{x,y+dy}})
		    write(s[ci..l])
		    x=temp+slen({s[ci]})
		else            -- scroll right end further right
		    p=l
		    while len>w do
			p=p-1
			len=slen(s[fvc..p]) end while
		    polygon(pc,1,{{sx,y},{sx+w-1,y},{sx+w-1,y+dy},{sx,y+dy}})
		    x=sx
		    write(s[fvc..p])
		    x=temp+slen({s[ci]}) end if
		ci=ci+1
	    end if
	end if
    end while
    -- restore attributes, justification, orientation & background
    at=att
    bo=pa[1]*and_bits(at,BOLD)
    italic=and_bits(at,ITALIC)>0
    ju=jus
    orient=ori
    bc=bco
    return {key,s}
end function   -- prompt

global procedure setx(integer i)
    -- set horizontal position of text pointer
    x=i
end procedure

global procedure sety(integer i)
    -- set vertical position of text pointer
    y=i
end procedure

global function getx()
    -- return current horizontal position of text pointer
    return x
end function

global function gety()
    -- return current vertical position of text pointer
    return y
end function

global function baseline_height()
    -- return baseline height of current font, in pixels
    return bh
end function

global function background()
    -- return current background color (CLEAR = -1)
    return bc
end function

global function foreground()
    -- return current foreground object: plain color or 2-d pattern
    return fc
end function

global procedure set_foreground(object c)
    -- set text foreground: plain color or 2-d pattern
    fc=c
end procedure

global function shadow_color()
    -- return current shadow color
    return sc
end function

global procedure set_shadow_color(integer c)
    -- set text shadow color
    sc=c
end procedure

global procedure set_background(integer c)
    -- set text background color
    bc=c
end procedure

global function orientation()
    -- return current line orientation
    return orient
end function

global procedure set_orientation(integer o)
    -- set line orientation, horizontal or vertical
    orient = o
end procedure

global function justification()
    -- return current line justification
    return ju
end function

global procedure justify(integer j)
    -- set line justification
    ju=j
end procedure

global procedure set_parameter(object n, object v)
    -- if n is atom the set n-th parameter of current font to value v
    -- else set entire sequence of parameters to corresponding values in v
    if atom(n) then
	pa[n]=v
	if n=1 then
	    bo=and_bits(BOLD,at)*pa[1] -- new bold offset
	end if
    else
	for i=1 to length(n) do
	    pa[n[i]]=v[i]
	    if n[i]=1 then
		bo=and_bits(BOLD,at)*pa[1] -- new bold offset
	    end if
	end for
    end if
end procedure

global function parameters()
    -- return sequence of current font parameters
    return pa
end function


-- init font system
load_rom_font()
select_font(ROM)
reset_font()

